import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:flutterx_http/src/x_http_client.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';

typedef RequestFactory = Future<Response> Function(XHttpClient client, JsonObject params, CancellationToken? cancel);

typedef ResponseConverter<T> = Future<T> Function(Response response);

@immutable
class ApiCall<T> {
  final RequestFactory _request;
  final ResponseConverter<T> _convert;

  const ApiCall._(this._request, this._convert);

  static ApiCall<void> empty({required RequestFactory request}) =>
      ApiCall._(request, (response) => SynchronousFuture(null));

  static ApiCall<Uint8List> bytes({required RequestFactory request}) =>
      ApiCall._(request, (response) => SynchronousFuture(response.bodyBytes));

  static ApiCall<EncodedData> bytesData({required RequestFactory request}) =>
      ApiCall._(request, (response) => SynchronousFuture(EncodedData(response.bodyBytes, response.encoding)));

  static ApiCall<String> string({required RequestFactory request}) =>
      ApiCall._(request, (response) => SynchronousFuture(response.body));

  static ApiCall<T> json<T>({required RequestFactory request, required FromJson<T> fromJson}) =>
      ApiCall._(request, (response) => SynchronousFuture(fromJson(jsonDecode(response.body))));

  static ApiCall<T> object<T>({required RequestFactory request, required FromJsonObject<T> fromJson}) =>
      ApiCall.json(request: request, fromJson: (json) => fromJson(json));

  static ApiCall<List<T>> list<T>({required RequestFactory request, required FromJsonObject<T> fromJson}) =>
      ApiCall.json(request: request, fromJson: (json) => fromJsonList(json, fromJson));

  static ApiCall<Map<String, T>> map<T>({required RequestFactory request, required FromJsonObject<T> fromJson}) =>
      ApiCall.json(request: request, fromJson: (json) => fromJsonMap(json, fromJson));

  Future<T> call(XHttpClient client, {JsonObject params = const {}, CancellationToken? cancel}) =>
      _request(client, params, cancel).then(_convert);

  /// Add this call after your API declaration to mock it by returning a mock [response]
  /// The returned value must be the body string or bytes
  /// You can also:
  /// - Set a [delay] to simulate server latency
  /// - Make the mock response dynamic by using query parameters passed to [response]
  /// - Throw an [XHTTPException] inside [response] to simulate errors
  ApiCall<T> mock({Duration delay = const Duration(seconds: 1), required FromJsonObject response}) =>
      _MockApiCall<T>(delay, response, _convert);
}

@immutable
class _MockApiCall<T> extends ApiCall<T> {
  final Duration _delay;
  final FromJsonObject _response;

  const _MockApiCall(this._delay, this._response, ResponseConverter<T> convert) : super._(_noRequest, convert);

  @override
  Future<T> call(XHttpClient client, {JsonObject params = const {}, CancellationToken? cancel}) {
    final raw = _response(params);
    final response = raw is List<int> ? Response.bytes(raw, 200) : Response(raw, 200);
    final result = Future.delayed(_delay, () => _convert(response));
    return cancel == null ? result : cancel.attach(result);
  }

  static Future<Response> _noRequest(XHttpClient client, JsonObject params, CancellationToken? cancel) =>
      throw UnimplementedError();
}

@immutable
class EncodedData {
  final Uint8List data;
  final Encoding encoding;

  const EncodedData(this.data, [this.encoding = utf8]);

  String decode() => encoding.decode(data);
}
