part of 'x_http_client.dart';

extension UriExt on Uri {
  /// Append [QueryParameters] to this [Uri]
  Uri appendQueryParameters(QueryParameters params) => Uri(
      scheme: scheme,
      userInfo: userInfo,
      host: host,
      port: port,
      path: path,
      queryParameters: Map<String, String>.from(queryParameters)
        ..addAll(params.map((key, value) => MapEntry(key, value.toString()))),
      fragment: hasFragment ? fragment : null);
}

extension BaseResponseExt on BaseResponse {
  bool get isSuccess => statusCode ~/ 100 == 2;

  Encoding get encoding => encodingForCharset(contentType.parameters['charset']);

  MediaType get contentType {
    final contentType = headers['content-type'];
    return contentType != null ? MediaType.parse(contentType) : MediaType('application', 'octet-stream');
  }
}

extension BaseRequestExt on BaseRequest {
  BaseRequest observeProgress(void Function(int readCount) onRead) => TransformRequest(
      source: this,
      transform: StreamTransformer.fromHandlers(
          handleData: (data, sink) {
            sink.add(data);
            onRead(data.length);
          },
          handleError: (error, stack, sink) => sink.addError(error, stack),
          handleDone: (sink) => sink.close()));

  TransformRequest transform({required StreamTransformer<List<int>, List<int>> transform}) =>
      TransformRequest(source: this, transform: transform);

  BaseRequest copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
  }) {
    if (this is Request)
      return (this as Request).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    if (this is StreamedRequest)
      return (this as StreamedRequest).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    if (this is MultipartRequest)
      return (this as MultipartRequest).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    return (this as CloneableRequest).copy(
        method: method,
        url: url,
        persistentConnection: persistentConnection,
        followRedirects: followRedirects,
        maxRedirects: maxRedirects,
        headers: headers);
  }
}

extension RequestExt on Request {
  Request copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
    String? body,
    List<int>? bodyBytes,
    Map<String, String>? bodyFields,
    Encoding? encoding,
  }) {
    final result = Request(method ?? this.method, url ?? this.url)
      ..persistentConnection = persistentConnection ?? this.persistentConnection
      ..followRedirects = followRedirects ?? this.followRedirects
      ..maxRedirects = maxRedirects ?? this.maxRedirects
      ..headers.addAll(headers ?? this.headers)
      ..encoding = encoding ?? this.encoding;
    if (body != null)
      result.body = body;
    else if (bodyBytes != null)
      result.bodyBytes = bodyBytes;
    else if (bodyFields != null)
      result.bodyFields = bodyFields;
    else
      result.bodyBytes = this.bodyBytes;
    return result;
  }
}

extension StreamedRequestExt on StreamedRequest {
  StreamedRequest copy({
    String? method,
    Uri? url,
    int? contentLength = -1,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
  }) =>
      StreamedRequest(method ?? this.method, url ?? this.url)
        ..contentLength = contentLength != -1 ? contentLength : this.contentLength
        ..persistentConnection = persistentConnection ?? this.persistentConnection
        ..followRedirects = followRedirects ?? this.followRedirects
        ..maxRedirects = maxRedirects ?? this.maxRedirects
        ..headers.addAll(headers ?? this.headers);
}

extension MultipartRequestExt on MultipartRequest {
  MultipartRequest copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
    Map<String, String>? fields,
    List<MultipartFile>? files,
  }) =>
      MultipartRequest(method ?? this.method, url ?? this.url)
        ..persistentConnection = persistentConnection ?? this.persistentConnection
        ..followRedirects = followRedirects ?? this.followRedirects
        ..maxRedirects = maxRedirects ?? this.maxRedirects
        ..headers.addAll(headers ?? this.headers)
        ..fields.addAll(fields ?? this.fields)
        ..files.addAll(files ?? this.files);
}

abstract class CloneableRequest {
  const CloneableRequest._();

  BaseRequest copy(
      {String? method,
      Uri? url,
      bool? persistentConnection,
      bool? followRedirects,
      int? maxRedirects,
      Map<String, String>? headers});
}
